import java.util.*;

public class ArboreBinar {
  private Nod radacina;

  /* CONSTRUCTORII SI METODELE CLASEI ArboreBinar */
  public ArboreBinar() {
    radacina=null;
  }

  public ArboreBinar(Nod radacina) {
    this.radacina=radacina;
  }

  public Nod radacina() {
    return radacina;
  }

  public Nod puneRadacina(Nod radacinaNoua) {
    Nod radacinaVeche=radacina;
    radacina=radacinaNoua;
    return radacinaVeche;
  }

  public int numarNoduri() {
    Iterator iter=iteratorLatime();
    int nrNoduri=0;
    while(iter.hasNext()) {
      iter.next();
      nrNoduri++;
    }
    return nrNoduri;
  }

  public Iterator iteratorAdancime() {
    return new ParcurgAdancime();
  }

  public Iterator iteratorPreordine() {
    return new TraversPreordine();
  }

  public Iterator iteratorInordine() {
    return new TraversInordine();
  }

  public Iterator iteratorPostordine() {
    return new TraversPostordine();
  }

  public Iterator iteratorLatime() {
    return new TraversLatime();
  }

  /* Metode recursive care intorc o lista cu informatiile din nodurile
     arborelui, respectand cele trei tehnici de traversare in adancime
  */

  /* Traversare in preordine */
  private void preordine(List lista, Nod nod) {
    lista.add(nod.info);
    if(nod.fiuStanga!=null) preordine(lista, nod.fiuStanga);
    if(nod.fiuDreapta!=null) preordine(lista, nod.fiuDreapta);
  }
  
  public List preordine() {
    ArrayList lista=new ArrayList();
    preordine(lista, radacina);
    return lista;
  }
    
  /* Traversare in inordine */
  private void inordine(List lista, Nod nod) {
    if(nod.fiuStanga!=null) inordine(lista, nod.fiuStanga);
    lista.add(nod.info);
    if(nod.fiuDreapta!=null) inordine(lista, nod.fiuDreapta);
  }
  
  public List inordine() {
    ArrayList lista=new ArrayList();
    inordine(lista, radacina);
    return lista;
  }

  /* Traversare in postordine */
  private void postordine(List lista, Nod nod) {
    if(nod.fiuStanga!=null) postordine(lista, nod.fiuStanga);
    if(nod.fiuDreapta!=null) postordine(lista, nod.fiuDreapta);
    lista.add(nod.info);
  }
  
  public List postordine() {
    ArrayList lista=new ArrayList();
    postordine(lista, radacina);
    return lista;
  }

  /* Metoda nerecursiva care intoarce lista informatiilor din noduri,
     traversate in latime. Foloseste iteratorul de traversare a
     arborelui in latime.
  */
  public List inLatime() {
    Iterator iter=iteratorLatime();
    List lista=new ArrayList();
    while(iter.hasNext()) lista.add(iter.next());
    return lista;
  }

  /* Metoda auxiliara recursiva care converteste un subarbore intr-un
     sir, in care arborele este pus in formatul cu paranteze
  */
  private static StringBuffer cuParanteze(StringBuffer sb, Nod nod) {
    sb.append("("+nod.info);
    if(nod.fiuStanga!=null) cuParanteze(sb, nod.fiuStanga);
    else if(nod.fiuDreapta!=null) sb.append("(()");
    if(nod.fiuDreapta!=null) cuParanteze(sb, nod.fiuDreapta);
    else if(nod.fiuStanga!=null) sb.append("()");
    sb.append(")");
    return sb;
  }

  /* Redefinirea metodei toString din clasa Object. Foloseste metoda
     auxiliara cuParanteze() aplicata radacinii arborelui
  */
  public String toString() {
    StringBuffer sb=new StringBuffer();
    if(radacina==null) return "( )";
    return cuParanteze(sb, radacina).toString();
  }

  /* CLASE INTERIOARE */

  public static class Nod {
    private Object info;
    private Nod fiuStanga, fiuDreapta;

    public Nod(Object obj) {
      info=obj;
      fiuStanga=fiuDreapta=null;
    }

    public Object info() {
      return info;
    }

    public Nod fiuStanga() {
      return fiuStanga;
    }

    public Nod fiuDreapta() {
      return fiuDreapta;
    }

    public Nod puneFiuStanga(Nod nod) {
      Nod deIntors=fiuStanga;
      fiuStanga=nod;
      return deIntors;
    }

    public Nod puneFiuDreapta(Nod nod) {
      Nod deIntors=fiuDreapta;
      fiuDreapta=nod;
      return deIntors;
    }
  
    public String toString() {
      return "["+info.toString()+"]";
    }
  }


  /* Clasa interioara privata ElemStiva reprezinta elementele stivei
     folosita ca structura auxiliara la parcurgerea arborelui in 
     adancime. Fiecare element contine referinta la un nod al
     arborelui si un indicator al ultimului fiu vizitat:
       0 - nici un fiu (prima trecere prin nod)
       1 - a fost vizitat fiuStanga
       2 - a fost vizitat fiuDreapta
  */
  private class ElemStiva {
    int fiuVizitat;
    Nod nod;

    ElemStiva(Nod nodCurent, int fiuVizitat) {
      this.fiuVizitat=fiuVizitat;
      nod=nodCurent;
    }

    public String toString() {
      return "<"+fiuVizitat+" "+nod+">";
    }
  }

  /* Clasa ParcurgAdancime defineste un iterator pentru parcurgerea
     arborelui in adancime. Foloseste ca structura auxiliara o stiva
     din clasa java.util.Stack. Elementele stivei sunt instante ale
     clasei interioare ElemStiva
  */
  public class ParcurgAdancime implements Iterator {
    Stack stiva;
    ParcurgAdancime() {
      stiva=new Stack();
      if(radacina!=null) stiva.push(new ElemStiva(radacina, 0));
    }
      
    public boolean hasNext() {
      if(stiva.empty()) return false;
      return true;
    }

    public Object next() {
      if(stiva.empty()) return null;
      ElemStiva varfStiva=(ElemStiva)stiva.peek();
      Nod nodCurent=varfStiva.nod;
      int fiuVizitat=varfStiva.fiuVizitat;
      if(fiuVizitat==0) {
        if(nodCurent.fiuStanga!=null) {
          stiva.push(new ElemStiva(nodCurent.fiuStanga, 0));
          varfStiva.fiuVizitat=1;
        }
        else if(nodCurent.fiuDreapta!=null) {
          stiva.push(new ElemStiva(nodCurent.fiuDreapta, 0));
          varfStiva.fiuVizitat=2;
        }
        else stiva.pop();
      }
      else if(fiuVizitat==1) {
        if(nodCurent.fiuDreapta!=null) {
          stiva.push(new ElemStiva(nodCurent.fiuDreapta, 0));
          varfStiva.fiuVizitat=2;
        }
        else stiva.pop();
      } else stiva.pop();
      return new ElemStiva(nodCurent, fiuVizitat);
    }

    public void remove() {}


  }

  /* Clasa TraversPreordine defineste un iterator pentru traversarea
     arborelui in preordine. La randul lui, pentru parcurgerea
     arborelui in adancime, acesta foloseste un iterator din clasa
     interioara ParcurgAdancime
  */
  public class TraversPreordine implements Iterator {
    ParcurgAdancime parcurg;
     
    public TraversPreordine() {
      parcurg=(ParcurgAdancime)iteratorAdancime();
    }
            
    public boolean hasNext() {
      while(parcurg.hasNext() && !parcurg.stiva.empty() && 
                ((ElemStiva)parcurg.stiva.peek()).fiuVizitat>0)
        parcurg.next();
      return parcurg.hasNext();
    }

    public Object next() {
      if(parcurg.hasNext()) {
        ElemStiva urmator=(ElemStiva)parcurg.next();
        return urmator.nod.info;
      }
      return null;
    }

    public void remove() {}
  }

  /* Clasa TraversInordine defineste un iterator pentru traversarea
     arborelui in inordine. Pentru parcurgerea in adancime a arborelui
     se foloseste un iterator din clasa interioara ParcurgAdancime
  */
  public class TraversInordine implements Iterator {
    ParcurgAdancime parcurg;
     
    public TraversInordine() {
      parcurg=(ParcurgAdancime)iteratorAdancime();
    }

    /* Se testeaza daca elementul curent (din varful stivei) poate fi
       vizitat in inordine
    */    
    private boolean vizitInordine() {
      if(parcurg.stiva.empty()) return false;
      ElemStiva curent=(ElemStiva)parcurg.stiva.peek();
      //if(curent==null) return false;
      if(curent.fiuVizitat==1) return true;
      if(curent.fiuVizitat==0 && curent.nod.fiuStanga==null)
                                              return true;
      return false;
    }
        
    public boolean hasNext() {
      if(!parcurg.hasNext()) return false;
      while(parcurg.hasNext() && !vizitInordine()) parcurg.next();
      return parcurg.hasNext();
    }

    public Object next() {
      if(parcurg.hasNext()) {
        ElemStiva urmator=(ElemStiva)parcurg.next();
        return urmator.nod.info;
      }
      return null;
    }

    public void remove() {}
  }

  /* Clasa TraversPostordine defineste un iterator pentru traversarea
     arborelui in postordine. Pentru parcurgerea in adancime a
     arborelui se foloseste un iterator din clasa interioara
     ParcurgAdancime
  */
  public class TraversPostordine implements Iterator {
    ParcurgAdancime parcurg;
     
    public TraversPostordine() {
      parcurg=(ParcurgAdancime)iteratorAdancime();
    }

    /* Se testeaza daca elementul curent (din varful stivei) poate fi
       vizitat in postordine
    */    
    private boolean vizitPostordine() {
      if(parcurg.stiva.empty()) return false;
      ElemStiva curent=(ElemStiva)parcurg.stiva.peek();
      //if(curent==null) return false;
      if(curent.fiuVizitat==2) return true;
      if(curent.nod.fiuDreapta==null && (curent.fiuVizitat==1 ||
        curent.fiuVizitat==0 && curent.nod.fiuStanga==null))
                                              return true;
      return false;
    }
        
    public boolean hasNext() {
      if(!parcurg.hasNext()) return false;
      while(parcurg.hasNext() && !vizitPostordine()) parcurg.next();
      return parcurg.hasNext();
    }

    public Object next() {
      if(parcurg.hasNext()) {
        ElemStiva urmator=(ElemStiva)parcurg.next();
        return urmator.nod.info;
      }
      return null;
    }

    public void remove() {}
  }

  /* Clasa TraversLatime este un iterator pentru traversarea arborelui
     in latime. Foloseste ca structura auxiliara o coada implementata
     ca instanta a clasei java.util.LinkedList
  */
     
  public class TraversLatime implements Iterator {
    LinkedList coada;

    public TraversLatime() {
      coada=new LinkedList();
      if(radacina!=null) coada.addLast(radacina);
    }

    public boolean hasNext() {
      if(coada.size()==0) return false;
      return true;
    }

    public Object next() {
      /* Se extrage din coada nodul curent */
      Nod nodCurent=(Nod)coada.removeFirst();
      /* Se pun in coada fiii nodului curent */
      if(nodCurent.fiuStanga!=null) 
            coada.addLast(nodCurent.fiuStanga);
      if(nodCurent.fiuDreapta!=null) 
            coada.addLast(nodCurent.fiuDreapta);
      /* Se intoarce informatia din nodul curent */
      return nodCurent.info;
    }

    public void remove() {}

    public Nod nodCurent() {
      return (Nod)coada.getFirst();
    }
  }
}
  