import java.util.*;

public class Arbore {
  private Nod radacina;

  /* CONSTRUCTORI SI METODE ALE CLASEI Arbore */ 

  public Arbore() {
    radacina=null;
  }

  public Arbore(Nod radacina) {
    this.radacina=radacina;
  }

  public Nod radacina() {
    return radacina;
  }

  public int numarNoduri() {
    Iterator iter=iteratorLatime();
    int nrNoduri=0;
    while(iter.hasNext()) {
      iter.next();
      nrNoduri++;
    }
    return nrNoduri;
  }

  public Nod puneRadacina(Nod radacinaNoua) {
    Nod radacinaVeche=radacina;
    radacina=radacinaNoua;
    return radacinaVeche;
  }

  public Iterator iteratorAdancime() {
    return new ParcurgAdancime();
  }

  public Iterator iteratorPreordine() {
    return new TraversPreordine();
  }

  public Iterator iteratorPostordine() {
    return new TraversPostordine();
  }

  public Iterator iteratorLatime() {
    return new TraversLatime();
  }

  /* Metode auxiliare recursive de creare a unor liste, care contin 
     informatiile din nodurile arborelui traversat in preordine,
     sau postordine.
  */

  /* Traversarea in preordine */
  private static void preordine(List lista, Nod nod) {
    lista.add(nod.info);
    Iterator iter=nod.fii.iterator();
    while(iter.hasNext())
      preordine(lista, (Nod)iter.next());
  }

  public List preordine() {
    List lista=new ArrayList();
    preordine(lista, radacina);
    return lista;
  }

  /* Traversarea in postordine */
  private static void postordine(List lista, Nod nod) {
    Iterator iter=nod.fii.iterator();
    while(iter.hasNext())
      postordine(lista, (Nod)iter.next());
    lista.add(nod.info);
  }

  public List postordine() {
    List lista=new ArrayList();
    postordine(lista, radacina);
    return lista;
  }

  /* Metoda nerecursiva care intoarce o lista cu informatiile din
     nodurile arborelui, traversat in latime
  */
  public List inLatime() {
    List lista=new ArrayList();
    Iterator iter=iteratorLatime();
    while(iter.hasNext()) 
      lista.add(iter.next());
    return lista;
  }

  /* Metoda auxiliara recursiva care converteste un subarbore intr-un
     sir, in care arborele este pus in formatul cu paranteze
  */
  private static StringBuffer cuParanteze(StringBuffer sb, Nod nod) {
    sb.append("("+nod);
    Iterator iter=nod.fii.iterator();
    while(iter.hasNext()) cuParanteze(sb, (Nod)iter.next());
    sb.append(")");
    return sb;
  }

  /* Redefinirea metodei toString din clasa Object. Foloseste metoda
     auxiliara cuParanteze() aplicata radacinii arborelui
  */
  public String toString() {
    StringBuffer sb=new StringBuffer();
    if(radacina==null) return "( )";
    return cuParanteze(sb, radacina).toString();
  }

  /* CLASE INTERIOARE */

  public static class Nod {
    private Object info;
    private ArrayList fii;

    public Nod(Object obj) {
      info=obj;
      fii=new ArrayList();
    }

    public Object info() {
      return info();
    }

    public Object puneInfo(Object infoNoua) {
      Object infoVeche=info;
      info=infoNoua;
      return infoVeche;
    }

    public ArrayList fii() {
      return fii;
    }

    public void puneFiu(Nod fiu) {
      fii.add(fiu);
    }

    public String toString() {
      return "["+info+"]";
    }
  } // sfarsit clasa Nod

  /* Clasa ElemStiva are ca instante elementele stivei folosite
     la parcurgerea arborelui in adancime
  */
  private class ElemStiva {
    Nod nod;
    Iterator iteratorFii;
    int nrFiiParcursi; 

    ElemStiva(Nod nod) {
      this.nod=nod;
      iteratorFii=nod.fii.iterator();
      nrFiiParcursi=0; 
    } 

    ElemStiva(ElemStiva elem) {
      nod=elem.nod;
      iteratorFii=elem.iteratorFii;
      nrFiiParcursi=elem.nrFiiParcursi;
    }

    public String toString() {
      return "<"+nrFiiParcursi+" "+nod+">"; 
    }
  }

  /* Clasa ParcurgAdancime ofera metodele necesare pentru parcurgerea
     in adancime a arborelui. Foloseste ca structura auxiliara o stiva
     din clasa java.util.Stack. Elementele stivei sunt instante ale 
     clasei auxiliare interioare ElemStiva.
  */
  public class ParcurgAdancime implements Iterator {
    Stack stiva;

    ParcurgAdancime() {
      stiva=new Stack();
      if(radacina!=null) stiva.push(new ElemStiva(radacina));
    }

    public boolean hasNext() {
      if(stiva.empty()) return false;
      return true;
    }

    public Object next() {
      if(stiva.empty()) return null;
      ElemStiva elemCurent=(ElemStiva)stiva.peek();
      ElemStiva deIntors=new ElemStiva(elemCurent);
      if(elemCurent.iteratorFii.hasNext()) {
        elemCurent.nrFiiParcursi++;
        stiva.push(new ElemStiva((Nod)elemCurent.iteratorFii.next()));
      } else stiva.pop();
      return deIntors;
    }

    public void remove() {}

  }

  /* Clasa TraversPreordine defineste un iterator pentru traversarea
     arborelui in preordine. La randul lui, pentru parcurgerea
     arborelui in adancime, acesta foloseste un iterator din clasa
     interioara ParcurgAdancime
  */
  public class TraversPreordine implements Iterator {
    ParcurgAdancime parcurg;
     
    public TraversPreordine() {
      parcurg=(ParcurgAdancime)iteratorAdancime();
    }
            
    public boolean hasNext() {
      while(parcurg.hasNext() && !parcurg.stiva.empty() && 
                ((ElemStiva)parcurg.stiva.peek()).nrFiiParcursi!=0)
        parcurg.next();
      return parcurg.hasNext();
    }

    public Object next() {
      if(parcurg.hasNext()) {
        ElemStiva urmator=(ElemStiva)parcurg.next();
        return urmator.nod.info;
      }
      return null;
    }

    public void remove() {}
  }
  
  /* Clasa TraversPostordine defineste un iterator pentru traversarea
     arborelui in postordine. Pentru parcurgerea in adancime a
     arborelui se foloseste un iterator din clasa interioara
     ParcurgAdancime
  */
  public class TraversPostordine implements Iterator {
    ParcurgAdancime parcurg;
     
    public TraversPostordine() {
      parcurg=(ParcurgAdancime)iteratorAdancime();
    }

    /* Se testeaza daca elementul curent (din varful stivei) poate fi
       vizitat in postordine
    */    
    private boolean vizitPostordine() {
      if(parcurg.stiva.empty()) return false;
      ElemStiva curent=(ElemStiva)parcurg.stiva.peek();
      if(curent.nrFiiParcursi==curent.nod.fii.size())
        return true;
      return false;
    }
        
    public boolean hasNext() {
      while(parcurg.hasNext() && !vizitPostordine()) parcurg.next();
      return parcurg.hasNext();
    }

    public Object next() {
      if(parcurg.hasNext()) {
        ElemStiva urmator=(ElemStiva)parcurg.next();
        return urmator.nod.info;
      }
      return null;
    }

    public void remove() {}
  }

  /* Clasa TraversLatime este un iterator pentru traversarea arborelui
     in latime. Foloseste ca structura auxiliara o coada implementata
     ca instanta a clasei java.util.LinkedList
  */
     
  public class TraversLatime implements Iterator {
    LinkedList coada;

    public TraversLatime() {
      coada=new LinkedList();
      if(radacina!=null) coada.addLast(radacina);
    }

    public boolean hasNext() {
      if(coada.size()==0) return false;
      return true;
    }

    public Object next() {
      /* Se extrage din coada nodul curent */
      Nod nodCurent=(Nod)coada.removeFirst();
      /* Se pun in coada fiii nodului curent */
      Iterator iter=nodCurent.fii.iterator();
      while(iter.hasNext()) coada.addLast(iter.next());
      /* Se intoarce informatia din nodul curent */
      return nodCurent.info;
    }

    public void remove() {}

    public Nod nodCurent() {
      return (Nod)coada.getFirst();
    }
  }
}
